package dev.fyloz.colorrecipesexplorer.exception

import com.fasterxml.jackson.annotation.JsonProperty
import dev.fyloz.colorrecipesexplorer.model.Material
import dev.fyloz.colorrecipesexplorer.model.MaterialQuantityDto
import org.springframework.context.annotation.Profile
import org.springframework.http.HttpHeaders
import org.springframework.http.HttpStatus
import org.springframework.http.ResponseEntity
import org.springframework.validation.FieldError
import org.springframework.web.bind.MethodArgumentNotValidException
import org.springframework.web.bind.annotation.ControllerAdvice
import org.springframework.web.bind.annotation.ExceptionHandler
import org.springframework.web.context.request.WebRequest
import org.springframework.web.servlet.mvc.method.annotation.ResponseEntityExceptionHandler

abstract class RestException(val exceptionMessage: String, val httpStatus: HttpStatus) :
    RuntimeException(exceptionMessage) {
    abstract fun buildBody(): RestExceptionBody

    @Suppress("unused")
    open inner class RestExceptionBody(
        val status: Int = httpStatus.value(),
        @JsonProperty("message") val message: String = exceptionMessage
    )
}

class EntityAlreadyExistsException(val value: Any) :
    RestException("An entity with the given identifier already exists", HttpStatus.CONFLICT) {
    @Suppress("unused")
    override fun buildBody(): RestExceptionBody = object : RestExceptionBody() {
        val id = value
    }
}

class EntityNotFoundException(val value: Any) :
    RestException("An entity could not be found with the given identifier", HttpStatus.NOT_FOUND) {
    @Suppress("unused")
    override fun buildBody(): RestExceptionBody = object : RestExceptionBody() {
        val id = value
    }
}

class CannotDeleteEntityException(val value: Long) :
    RestException(
        "The entity with the given identifier could not be deleted because it is required by other entities",
        HttpStatus.CONFLICT
    ) {
    @Suppress("unused")
    override fun buildBody(): RestExceptionBody = object : RestExceptionBody() {
        val id = value
    }
}

class SimdutWriteException(val material: Material) :
    RestException(
        "Could not write the SIMDUT file to disk",
        HttpStatus.INTERNAL_SERVER_ERROR
    ) {
    @Suppress("unused")
    override fun buildBody(): RestExceptionBody = RestExceptionBody()
}

class LowQuantityException(val materialQuantity: MaterialQuantityDto) :
    RestException(
        "There is not enough of the given material in the inventory",
        HttpStatus.CONFLICT
    ) {
    @Suppress("unused")
    override fun buildBody(): RestExceptionBody = object : RestExceptionBody() {
        val material = materialQuantity.material
        val quantity = materialQuantity.quantity
    }
}

class LowQuantitiesException(val materialQuantities: Collection<MaterialQuantityDto>) :
    RestException(
        "There is not enough of one or more given materials in the inventory",
        HttpStatus.CONFLICT
    ) {
    @Suppress
    override fun buildBody(): RestExceptionBody = object : RestExceptionBody() {
        val lowQuantities = materialQuantities
    }
}

@ControllerAdvice
class RestResponseEntityExceptionHandler : ResponseEntityExceptionHandler() {
    @ExceptionHandler(RestException::class)
    fun handleRestExceptions(exception: RestException, request: WebRequest): ResponseEntity<Any> {
        return handleExceptionInternal(exception, exception.buildBody(), HttpHeaders(), exception.httpStatus, request)
    }

    override fun handleMethodArgumentNotValid(
        ex: MethodArgumentNotValidException,
        headers: HttpHeaders,
        status: HttpStatus,
        request: WebRequest
    ): ResponseEntity<Any> {
        val errors = hashMapOf<String, String>()
        ex.bindingResult.allErrors.forEach {
            val fieldName = (it as FieldError).field
            val errorMessage = it.defaultMessage!!
            errors[fieldName] = errorMessage
        }
        return ResponseEntity(errors, headers, status)
    }
}
