package dev.fyloz.colorrecipesexplorer.model

import com.fasterxml.jackson.annotation.JsonIgnore
import com.fasterxml.jackson.annotation.JsonProperty
import dev.fyloz.colorrecipesexplorer.model.validation.NullOrNotBlank
import org.hibernate.annotations.Fetch
import org.hibernate.annotations.FetchMode
import org.springframework.security.core.GrantedAuthority
import org.springframework.security.core.authority.SimpleGrantedAuthority
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder
import org.springframework.security.crypto.password.PasswordEncoder
import java.time.LocalDateTime
import javax.persistence.*
import javax.validation.constraints.NotBlank
import javax.validation.constraints.NotNull
import javax.validation.constraints.Size


private const val EMPLOYEE_ID_NULL_MESSAGE = "Un numéro d'employé est requis"
private const val EMPLOYEE_LAST_NAME_EMPTY_MESSAGE = "Un nom  est requis"
private const val EMPLOYEE_FIRST_NAME_EMPTY_MESSAGE = "Un prénom est requis"
private const val EMPLOYEE_PASSWORD_EMPTY_MESSAGE = "Un mot de passe est requis"
private const val EMPLOYEE_PASSWORD_TOO_SHORT_MESSAGE = "Le mot de passe doit contenir au moins 8 caractères"

@Entity
@Table(name = "employee")
data class Employee(
    @Id
    override val id: Long,

    @Column(name = "first_name")
    val firstName: String = "",

    @Column(name = "last_name")
    val lastName: String = "",

    @JsonIgnore
    val password: String = "",

    @JsonIgnore
    @Column(name = "default_group_user")
    val isDefaultGroupUser: Boolean = false,

    @JsonIgnore
    @Column(name = "system_user")
    val isSystemUser: Boolean = false,

    @ManyToOne
    @JoinColumn(name = "group_id")
    @Fetch(FetchMode.SELECT)
    var group: EmployeeGroup? = null,

    @Enumerated(EnumType.STRING)
    @ElementCollection(fetch = FetchType.EAGER)
    @CollectionTable(name = "employee_permission", joinColumns = [JoinColumn(name = "employee_id")])
    @Column(name = "permission")
    @Fetch(FetchMode.SUBSELECT)
    @get:JsonProperty("explicitPermissions")
    val permissions: MutableSet<EmployeePermission> = mutableSetOf(),

    @Column(name = "last_login_time")
    var lastLoginTime: LocalDateTime? = null
) : Model {
    @get:JsonProperty("permissions")
    val flatPermissions: Set<EmployeePermission>
        get() = permissions
            .flatMap { it.flat() }
            .filter { !it.deprecated }
            .toMutableSet()
            .apply {
                if (group != null) this.addAll(group!!.flatPermissions)
            }

    @get:JsonIgnore
    val authorities: Set<GrantedAuthority>
        get() = flatPermissions.map { it.toAuthority() }.toMutableSet()
}

/** DTO for creating employees. Allows a [password] a [groupId]. */
open class EmployeeSaveDto(
    @field:NotNull(message = EMPLOYEE_ID_NULL_MESSAGE)
    val id: Long,

    @field:NotBlank(message = EMPLOYEE_FIRST_NAME_EMPTY_MESSAGE)
    val firstName: String,

    @field:NotBlank(message = EMPLOYEE_LAST_NAME_EMPTY_MESSAGE)
    val lastName: String,

    @field:NotBlank(message = EMPLOYEE_PASSWORD_EMPTY_MESSAGE)
    @field:Size(min = 8, message = EMPLOYEE_PASSWORD_TOO_SHORT_MESSAGE)
    val password: String,

    val groupId: Long?,

    @Enumerated(EnumType.STRING)
    val permissions: MutableSet<EmployeePermission> = mutableSetOf()
) : EntityDto<Employee>

open class EmployeeUpdateDto(
    @field:NotNull(message = EMPLOYEE_ID_NULL_MESSAGE)
    val id: Long,

    @field:NullOrNotBlank(message = EMPLOYEE_FIRST_NAME_EMPTY_MESSAGE)
    val firstName: String?,

    @field:NullOrNotBlank(message = EMPLOYEE_LAST_NAME_EMPTY_MESSAGE)
    val lastName: String?,

    val groupId: Long?,

    @Enumerated(EnumType.STRING)
    val permissions: Set<EmployeePermission>?
) : EntityDto<Employee>


private const val GROUP_ID_NULL_MESSAGE = "Un identifiant est requis"
private const val GROUP_NAME_NULL_MESSAGE = "Un nom est requis"
private const val GROUP_PERMISSIONS_EMPTY_MESSAGE = "Au moins une permission est requise"

@Entity
@Table(name = "employee_group")
data class EmployeeGroup(
    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    override var id: Long? = null,

    @Column(unique = true)
    override val name: String = "",

    @Enumerated(EnumType.STRING)
    @ElementCollection(fetch = FetchType.EAGER)
    @CollectionTable(name = "group_permission", joinColumns = [JoinColumn(name = "group_id")])
    @Column(name = "permission")
    @Fetch(FetchMode.SUBSELECT)
    @get:JsonProperty("explicitPermissions")
    val permissions: MutableSet<EmployeePermission> = mutableSetOf(),
) : NamedModel {
    @get:JsonProperty("permissions")
    val flatPermissions: Set<EmployeePermission>
        get() = this.permissions
            .flatMap { it.flat() }
            .filter { !it.deprecated }
            .toSet()
}

open class EmployeeGroupSaveDto(
    @field:NotBlank(message = GROUP_NAME_NULL_MESSAGE)
    @field:Size(min = 3)
    val name: String,

    @field:Size(min = 1, message = GROUP_PERMISSIONS_EMPTY_MESSAGE)
    val permissions: MutableSet<EmployeePermission>
) : EntityDto<EmployeeGroup> {
    override fun toEntity(): EmployeeGroup =
        EmployeeGroup(null, name, permissions)
}

open class EmployeeGroupUpdateDto(
    @field:NotNull(message = GROUP_ID_NULL_MESSAGE)
    val id: Long,

    @field:NotBlank(message = GROUP_NAME_NULL_MESSAGE)
    @field:Size(min = 3)
    val name: String,

    @field:Size(min = 1, message = GROUP_PERMISSIONS_EMPTY_MESSAGE)
    val permissions: MutableSet<EmployeePermission>
) : EntityDto<EmployeeGroup> {
    override fun toEntity(): EmployeeGroup =
        EmployeeGroup(id, name, permissions)
}


data class EmployeeLoginRequest(val id: Long, val password: String)


enum class EmployeePermission(
    val impliedPermissions: List<EmployeePermission> = listOf(),
    val deprecated: Boolean = false
) {
    VIEW_RECIPES,
    VIEW_CATALOG,
    VIEW_USERS,

    PRINT_MIXES(listOf(VIEW_RECIPES)),

    EDIT_RECIPES_PUBLIC_DATA(listOf(VIEW_RECIPES)),
    EDIT_RECIPES(listOf(EDIT_RECIPES_PUBLIC_DATA)),
    EDIT_MATERIALS(listOf(VIEW_CATALOG)),
    EDIT_MATERIAL_TYPES(listOf(VIEW_CATALOG)),
    EDIT_COMPANIES(listOf(VIEW_CATALOG)),
    EDIT_USERS(listOf(VIEW_USERS)),
    EDIT_CATALOG(listOf(EDIT_MATERIALS, EDIT_MATERIAL_TYPES, EDIT_COMPANIES)),

    ADD_TO_INVENTORY(listOf(VIEW_CATALOG)),
    DEDUCT_FROM_INVENTORY(listOf(VIEW_RECIPES)),

    REMOVE_RECIPES(listOf(EDIT_RECIPES)),
    REMOVE_MATERIALS(listOf(EDIT_MATERIALS)),
    REMOVE_MATERIAL_TYPES(listOf(EDIT_MATERIAL_TYPES)),
    REMOVE_COMPANIES(listOf(EDIT_COMPANIES)),
    REMOVE_USERS(listOf(EDIT_USERS)),
    REMOVE_CATALOG(listOf(REMOVE_MATERIALS, REMOVE_MATERIAL_TYPES, REMOVE_COMPANIES)),

    ADMIN(
        listOf(
            REMOVE_RECIPES,
            REMOVE_USERS,
            REMOVE_CATALOG,

            PRINT_MIXES,
            ADD_TO_INVENTORY,
            DEDUCT_FROM_INVENTORY
        )
    ),

    // deprecated permissions
    VIEW_RECIPE(listOf(VIEW_RECIPES), true),
    VIEW_MATERIAL(listOf(VIEW_CATALOG), true),
    VIEW_MATERIAL_TYPE(listOf(VIEW_CATALOG), true),
    VIEW_COMPANY(listOf(VIEW_CATALOG), true),
    VIEW(listOf(VIEW_RECIPES, VIEW_CATALOG), true),
    VIEW_EMPLOYEE(listOf(VIEW_USERS), true),
    VIEW_EMPLOYEE_GROUP(listOf(VIEW_USERS), true),

    EDIT_RECIPE(listOf(EDIT_RECIPES), true),
    EDIT_MATERIAL(listOf(EDIT_MATERIALS), true),
    EDIT_MATERIAL_TYPE(listOf(EDIT_MATERIAL_TYPES), true),
    EDIT_COMPANY(listOf(EDIT_COMPANIES), true),
    EDIT(listOf(EDIT_RECIPES, EDIT_CATALOG), true),
    EDIT_EMPLOYEE(listOf(EDIT_USERS), true),
    EDIT_EMPLOYEE_PASSWORD(listOf(EDIT_USERS), true),
    EDIT_EMPLOYEE_GROUP(listOf(EDIT_USERS), true),

    REMOVE_RECIPE(listOf(REMOVE_RECIPES), true),
    REMOVE_MATERIAL(listOf(REMOVE_MATERIALS), true),
    REMOVE_MATERIAL_TYPE(listOf(REMOVE_MATERIAL_TYPES), true),
    REMOVE_COMPANY(listOf(REMOVE_COMPANIES), true),
    REMOVE(listOf(REMOVE_RECIPES, REMOVE_CATALOG), true),
    REMOVE_EMPLOYEE(listOf(REMOVE_USERS), true),
    REMOVE_EMPLOYEE_GROUP(listOf(REMOVE_USERS), true),

    SET_BROWSER_DEFAULT_GROUP(listOf(VIEW_USERS), true),
    ;

    operator fun contains(permission: EmployeePermission): Boolean {
        return permission == this || impliedPermissions.any { permission in it }
    }
}

fun EmployeePermission.flat(): Iterable<EmployeePermission> {
    return mutableSetOf(this).apply {
        impliedPermissions.forEach {
            addAll(it.flat())
        }
    }
}

/** Converts the given [EmployeePermission] to a [GrantedAuthority]. */
private fun EmployeePermission.toAuthority(): GrantedAuthority {
    return SimpleGrantedAuthority(name)
}


// ==== DSL ====
fun employee(
    passwordEncoder: PasswordEncoder = BCryptPasswordEncoder(),
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    password: String = passwordEncoder.encode("password"),
    isDefaultGroupUser: Boolean = false,
    isSystemUser: Boolean = false,
    group: EmployeeGroup? = null,
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    lastLoginTime: LocalDateTime? = null,
    op: Employee.() -> Unit = {}
) = Employee(
    id,
    firstName,
    lastName,
    password,
    isDefaultGroupUser,
    isSystemUser,
    group,
    permissions,
    lastLoginTime
).apply(op)

fun employeeSaveDto(
    passwordEncoder: PasswordEncoder = BCryptPasswordEncoder(),
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    password: String = passwordEncoder.encode("password"),
    groupId: Long? = null,
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeSaveDto.() -> Unit = {}
) = EmployeeSaveDto(id, firstName, lastName, password, groupId, permissions).apply(op)

fun employeeUpdateDto(
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    groupId: Long? = null,
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeUpdateDto.() -> Unit = {}
) = EmployeeUpdateDto(id, firstName, lastName, groupId, permissions).apply(op)

fun employeeGroup(
    id: Long? = null,
    name: String = "name",
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeGroup.() -> Unit = {}
) = EmployeeGroup(id, name, permissions).apply(op)

fun employeeGroupSaveDto(
    name: String = "name",
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeGroupSaveDto.() -> Unit = {}
) = EmployeeGroupSaveDto(name, permissions).apply(op)

fun employeeGroupUpdateDto(
    id: Long = 0L,
    name: String = "name",
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeGroupUpdateDto.() -> Unit = {}
) = EmployeeGroupUpdateDto(id, name, permissions).apply(op)
