package dev.fyloz.colorrecipesexplorer.rest

import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeEditUsers
import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeRemoveUsers
import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeViewUsers
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.service.EmployeeGroupServiceImpl
import dev.fyloz.colorrecipesexplorer.service.EmployeeService
import org.springframework.http.MediaType
import org.springframework.security.access.prepost.PreAuthorize
import org.springframework.web.bind.annotation.*
import java.security.Principal
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import javax.validation.Valid

private const val EMPLOYEE_CONTROLLER_PATH = "api/employee"
private const val EMPLOYEE_GROUP_CONTROLLER_PATH = "api/employee/group"

@RestController
@RequestMapping(EMPLOYEE_CONTROLLER_PATH)
class EmployeeController(private val employeeService: EmployeeService) {
    @GetMapping
    @PreAuthorizeViewUsers
    fun getAll() =
        ok(employeeService.getAll())

    @GetMapping("{id}")
    @PreAuthorizeViewUsers
    fun getById(@PathVariable id: Long) =
        ok(employeeService.getById(id))

    @GetMapping("current")
    fun getCurrent(loggedInEmployee: Principal?) =
        if (loggedInEmployee != null)
            ok(
                employeeService.getById(
                    loggedInEmployee.name.toLong(),
                    ignoreDefaultGroupUsers = false,
                    ignoreSystemUsers = false
                )
            )
        else
            forbidden()

    @PostMapping
    @PreAuthorizeEditUsers
    fun save(@Valid @RequestBody employee: EmployeeSaveDto) =
        created<Employee>(EMPLOYEE_CONTROLLER_PATH) {
            employeeService.save(employee)
        }

    @PutMapping
    @PreAuthorizeEditUsers
    fun update(@Valid @RequestBody employee: EmployeeUpdateDto) =
        noContent {
            employeeService.update(employee)
        }

    @PutMapping("{id}/password", consumes = [MediaType.TEXT_PLAIN_VALUE])
    @PreAuthorizeEditUsers
    fun updatePassword(@PathVariable id: Long, @RequestBody password: String) =
        noContent {
            employeeService.updatePassword(id, password)
        }

    @PutMapping("{employeeId}/permissions/{permission}")
    @PreAuthorizeEditUsers
    fun addPermission(
        @PathVariable employeeId: Long,
        @PathVariable permission: EmployeePermission
    ) = noContent {
        employeeService.addPermission(employeeId, permission)
    }

    @DeleteMapping("{employeeId}/permissions/{permission}")
    @PreAuthorizeEditUsers
    fun removePermission(
        @PathVariable employeeId: Long,
        @PathVariable permission: EmployeePermission
    ) = noContent {
        employeeService.removePermission(employeeId, permission)
    }

    @DeleteMapping("{id}")
    @PreAuthorizeRemoveUsers
    fun deleteById(@PathVariable id: Long) =
        employeeService.deleteById(id)
}

@RestController
@RequestMapping(EMPLOYEE_GROUP_CONTROLLER_PATH)
@PreAuthorizeViewUsers
class GroupsController(private val groupService: EmployeeGroupServiceImpl) {
    @GetMapping
    fun getAll() =
        ok(groupService.getAll())

    @GetMapping("{id}")
    fun getById(@PathVariable id: Long) =
        ok(groupService.getById(id))

    @GetMapping("{id}/employees")
    fun getEmployeesForGroup(@PathVariable id: Long) =
        ok(groupService.getEmployeesForGroup(id))

    @PostMapping("default/{groupId}")
    fun setDefaultGroup(@PathVariable groupId: Long, response: HttpServletResponse) =
        noContent {
            groupService.setResponseDefaultGroup(groupId, response)
        }

    @GetMapping("default")
    fun getRequestDefaultGroup(request: HttpServletRequest) =
        ok(groupService.getRequestDefaultGroup(request))

    @PostMapping
    @PreAuthorizeEditUsers
    fun save(@Valid @RequestBody group: EmployeeGroupSaveDto) =
        created<EmployeeGroup>(EMPLOYEE_GROUP_CONTROLLER_PATH) {
            groupService.save(group)
        }

    @PutMapping
    @PreAuthorizeEditUsers
    fun update(@Valid @RequestBody group: EmployeeGroupUpdateDto) =
        noContent {
            groupService.update(group)
        }

    @DeleteMapping("{id}")
    @PreAuthorizeRemoveUsers
    fun deleteById(@PathVariable id: Long) =
        noContent {
            groupService.deleteById(id)
        }
}

@RestController
@RequestMapping("api")
class LogoutController(private val employeeService: EmployeeService) {
    @GetMapping("logout")
    fun logout(request: HttpServletRequest) =
        ok<Void> {
            employeeService.logout(request)
        }
}
