package dev.fyloz.colorrecipesexplorer.service.files

import dev.fyloz.colorrecipesexplorer.config.properties.CreProperties
import org.slf4j.Logger
import org.springframework.core.io.ResourceLoader
import org.springframework.stereotype.Service
import org.springframework.web.multipart.MultipartFile
import java.io.File
import java.io.IOException
import java.io.InputStream
import java.nio.charset.StandardCharsets
import java.nio.file.Files
import java.nio.file.Paths

@Service
class FileService(
    private val resourcesLoader: ResourceLoader,
    private val creProperties: CreProperties,
    private val logger: Logger
) {
    /** Reads the resource at the given [path] as a [String]. */
    fun readResource(path: String): String = try {
        resourcesLoader.getResource("classpath:$path").inputStream.use {
            readInputStreamAsString(it)
        }
    } catch (ex: IOException) {
        logger.error("Could not read resource", ex)
        ""
    }

    /** Reads the given [stream] as a [String]. */
    fun readInputStreamAsString(stream: InputStream) = with(stream.readAllBytes()) {
        String(this, StandardCharsets.UTF_8)
    }

    /** Reads the file at the given [path] as a [ByteArray]. */
    fun readAsBytes(path: String) =
        withFileAt(path) { this.readBytes() }

    /** Writes the given [multipartFile] to the file at the given [path]. */
    fun write(multipartFile: MultipartFile, path: String): Boolean =
        if (multipartFile.size <= 0) true
        else try {
            multipartFile.transferTo(create(path))
            true
        } catch (ex: IOException) {
            logger.error("Unable to write multipart file", ex)
            false
        }

    /** Creates a new file at the given [path]. If the file already exists, nothing will be done. */
    fun create(path: String) = withFileAt(path) {
        if (!exists(path)) {
            try {
                Files.createDirectories(this.parentFile.toPath())
                Files.createFile(this.toPath())
            } catch (ex: IOException) {
                logger.error("Unable to create file", ex)
            }
        }
        this
    }

    /** Deletes the file at the given [path]. */
    fun delete(path: String) = withFileAt(path) {
        try {
            if (exists(path)) Files.delete(this.toPath())
        } catch (ex: IOException) {
            logger.error("Unable to delete file", ex)
        }
    }

    /** Checks if a file with the given [path] exists on the disk. */
    fun exists(path: String): Boolean = withFileAt(path) {
        this.exists() && this.isFile
    }

    /** Runs the given [block] in the context of a file with the given [path]. */
    fun <T> withFileAt(path: String, block: File.() -> T) =
        File(path).block()

    fun getPath(fileName: String): String =
        "${creProperties.workingDirectory}/$fileName"
}
