package dev.fyloz.colorrecipesexplorer.service.files

import dev.fyloz.colorrecipesexplorer.exception.SimdutWriteException
import dev.fyloz.colorrecipesexplorer.model.Material
import dev.fyloz.colorrecipesexplorer.service.MaterialService
import org.slf4j.Logger
import org.springframework.context.annotation.Lazy
import org.springframework.stereotype.Service
import org.springframework.web.multipart.MultipartFile
import java.io.IOException

const val SIMDUT_DIRECTORY = "simdut"

@Service
class SimdutService(
    private val fileService: FileService,
    private val logger: Logger
) {
    /** Checks if the given [material] has a SIMDUT file. */
    fun exists(material: Material) =
        fileService.exists(getPath(material))

    /** Reads the SIMDUT file of the given [material]. */
    fun read(material: Material): ByteArray {
        val path = getPath(material)
        if (!fileService.exists(path)) return ByteArray(0)

        return try {
            fileService.readAsBytes(path)
        } catch (ex: IOException) {
            logger.error("Could not read SIMDUT file", ex)
            ByteArray(0)
        }
    }

    /** Writes the given [simdut] file for the given [material] to the disk. */
    fun write(material: Material, simdut: MultipartFile) {
        if (!fileService.write(simdut, getPath(material)))
            throw SimdutWriteException(material)
    }

    /** Updates the SIMDUT file of the given [material] with the given [simdut]. */
    fun update(simdut: MultipartFile, material: Material) {
        delete(material)
        write(material, simdut)
    }

    /** Deletes the SIMDUT file of the given [material]. */
    fun delete(material: Material) =
        fileService.delete(getPath(material))

    /** Gets the path of the SIMDUT file of the given [material]. */
    fun getPath(material: Material) =
        fileService.getPath("$SIMDUT_DIRECTORY/${getSimdutFileName(material)}")

    /** Gets the name of the SIMDUT file of the given [material]. */
    fun getSimdutFileName(material: Material) =
        material.id.toString()
}
