package dev.fyloz.colorrecipesexplorer.service

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.exception.EntityAlreadyExistsException
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.repository.MaterialRepository
import dev.fyloz.colorrecipesexplorer.service.files.SimdutService
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import org.springframework.mock.web.MockMultipartFile
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class MaterialServiceTest :
    AbstractExternalNamedModelServiceTest<Material, MaterialSaveDto, MaterialUpdateDto, MaterialService, MaterialRepository>() {
    override val repository: MaterialRepository = mock()
    private val simdutService: SimdutService = mock()
    private val recipeService: RecipeService = mock()
    private val mixService: MixService = mock()
    private val materialTypeService: MaterialTypeService = mock()
    override val service: MaterialService =
        spy(MaterialServiceImpl(repository, simdutService, recipeService, mixService, materialTypeService))

    override val entity: Material = material(id = 0L, name = "material")
    override val anotherEntity: Material = material(id = 1L, name = "another material")
    override val entityWithEntityName: Material = material(id = 2L, name = "material")
    override val entitySaveDto: MaterialSaveDto = spy(materialSaveDto())
    override val entityUpdateDto: MaterialUpdateDto = spy(materialUpdateDto(id = 0L))

    private val materialType = materialType()

    @AfterEach
    override fun afterEach() {
        reset(simdutService)
        super.afterEach()
    }

    // existsByMaterialType

    @Test
    fun `existsByMaterialType() returns true when a material with the given material type exists in the repository`() {
        whenever(repository.existsByMaterialType(materialType)).doReturn(true)

        val found = service.existsByMaterialType(materialType)

        assertTrue(found)
    }

    @Test
    fun `existsByMaterialType() returns false when no material with the given material type exists in the repository`() {
        whenever(repository.existsByMaterialType(materialType)).doReturn(false)

        val found = service.existsByMaterialType(materialType)

        assertFalse(found)
    }

    // hasSimdut()

    @Test
    fun `hasSimdut() returns false when simdutService_exists() returns false`() {
        whenever(simdutService.exists(entity)).doReturn(false)
        doReturn(entity).whenever(service).getById(entity.id!!)

        val found = service.hasSimdut(entity.id!!)

        assertFalse(found)
    }

    @Test
    fun `hasSimdut() returns true when simdutService_exists() returns true`() {
        whenever(simdutService.exists(entity)).doReturn(true)
        doReturn(entity).whenever(service).getById(entity.id!!)

        val found = service.hasSimdut(entity.id!!)

        assertTrue(found)
    }

    // getAllNotMixType()

    @Test
    fun `getAllNotMixType() returns a list containing every material that are not a mix type`() {
        val mixTypeMaterial = material(name = "mix type material", isMixType = true)
        val materialList = listOf(entity, mixTypeMaterial)

        doReturn(materialList).whenever(service).getAll()

        val found = service.getAllNotMixType()

        assertTrue(found.contains(entity))
        assertFalse(found.contains(mixTypeMaterial))
    }

    // getAllIdsWithSimdut()

    @Test
    fun `getAllIdsWithSimdut() returns a list containing the identifier of every material with a SIMDUT file`() {
        val materials = listOf(
            material(id = 0L),
            material(id = 1L),
            material(id = 2L),
            material(id = 3L)
        )
        val hasSimdut = mapOf(
            *materials
                .map { it.id!! to it.evenId }
                .toTypedArray()
        )
        val expectedIds = hasSimdut
            .filter { it.value }
            .map { it.key }

        whenever(simdutService.exists(any())).doAnswer { hasSimdut[(it.arguments[0] as Material).id] }
        doReturn(materials).whenever(service).getAllNotMixType()

        val found = service.getAllIdsWithSimdut()

        assertEquals(expectedIds, found)
    }

    // save()

    @Test
    fun `save() throws EntityAlreadyExistsException when a material with the given name exists in the repository`() {
        doReturn(true).whenever(service).existsByName(entity.name)

        val exception = assertThrows<EntityAlreadyExistsException> { service.save(entity) }
        assertEquals(entity.name, exception.value)
    }

    @Test
    override fun `save(dto) calls and returns save() with the created entity`() {
        withBaseSaveDtoTest(entity, entitySaveDto, service, { any() })
    }

    @Test
    fun `save(dto) calls simdutService_write() with the saved entity`() {
        val mockMultipartFile = spy(MockMultipartFile("simdut", byteArrayOf()))
        val materialSaveDto = spy(materialSaveDto(simdutFile = mockMultipartFile))

        doReturn(false).whenever(mockMultipartFile).isEmpty
        doReturn(entity).whenever(service).save(any<Material>())

        service.save(materialSaveDto)

        verify(simdutService).write(entity, mockMultipartFile)
    }

    // update()

    @Test
    fun `update() throws EntityAlreadyExistsException when another material with the updated name exists in the repository`() {
        val material = material(id = 0L, name = "name")
        val anotherMaterial = material(id = 1L, name = "name")

        whenever(repository.findByName(material.name)).doReturn(anotherMaterial)
        doReturn(entity).whenever(service).getById(material.id!!)

        val exception = assertThrows<EntityAlreadyExistsException> { service.update(material) }
        assertEquals(material.name, exception.value)
    }

    // updateQuantity()

    @Test
    fun `updateQuantity() updates the quantity of the the given material in the repository`() {
        val material = material(id = 0L, inventoryQuantity = 4321f)
        val quantity = 1234f
        val totalQuantity = material.inventoryQuantity + quantity

        val found = service.updateQuantity(material, quantity)

        verify(repository).updateInventoryQuantityById(material.id!!, totalQuantity)
        assertEquals(totalQuantity, found)
    }

    // getAllForMixCreation()

    @Test
    fun `getAllForMixCreation() returns all normal materials and all mix type materials for the given recipe`() {
        val normalMaterial = material(id = 0L, isMixType = false)
        val mixTypeMaterial = material(id = 1L, isMixType = true)
        val anotherMixTypeMaterial = material(id = 2L, isMixType = true)
        val materials = listOf(normalMaterial, mixTypeMaterial, anotherMixTypeMaterial)
        val recipe =
            recipe(id = 0L, mixes = mutableListOf(mix(mixType = mixType(id = 0L, material = mixTypeMaterial))))

        whenever(recipeService.getById(recipe.id!!)).doReturn(recipe)
        doReturn(materials).whenever(service).getAll()

        val found = service.getAllForMixCreation(recipe.id!!)

        assertTrue(normalMaterial in found)
        assertTrue(mixTypeMaterial in found)
        assertFalse(anotherMixTypeMaterial in found)
    }

    // getAllForMixUpdate()

    @Test
    fun `getAllForMixUpdate() returns all normal materials and all mix type materials for the recipe of the given mix without the mix type of the said mix`() {
        val normalMaterial = material(id = 0L, isMixType = false)
        val mixTypeMaterial = material(id = 1L, isMixType = true)
        val anotherMixTypeMaterial = material(id = 2L, isMixType = true)
        val materials = listOf(normalMaterial, mixTypeMaterial, anotherMixTypeMaterial)
        val recipe = recipe(id = 0L, mixes = mutableListOf(mix(mixType = mixType(material = mixTypeMaterial))))
        val mix = mix(id = 1L, recipe = recipe, mixType = mixType(material = anotherMixTypeMaterial))
        recipe.mixes.add(mix)

        whenever(mixService.getById(mix.id!!)).doReturn(mix)
        doReturn(materials).whenever(service).getAll()

        val found = service.getAllForMixUpdate(mix.id!!)

        assertTrue(normalMaterial in found)
        assertTrue(mixTypeMaterial in found)
        assertFalse(anotherMixTypeMaterial in found)
    }

    // update()

    @Test
    override fun `update(dto) calls and returns update() with the created entity`() {
        val mockSimdutFile = MockMultipartFile("simdut", byteArrayOf(1, 2, 3, 4, 5))
        val materialUpdateDto = spy(materialUpdateDto(id = 0L, simdutFile = mockSimdutFile))

        doReturn(entity).whenever(service).getById(materialUpdateDto.id)
        doReturn(entity).whenever(service).update(any<Material>())
        doReturn(entity).whenever(materialUpdateDto).toEntity()

        service.update(materialUpdateDto)

        verify(simdutService).update(eq(mockSimdutFile), any())
    }


    // delete()

    override fun `deleteById() deletes the entity with the given id in the repository`() {
        super.`deleteById() deletes the entity with the given id in the repository`()
    }

    /** Utility property to check if the identifier of the given [Material] is even. */
    val Material.evenId: Boolean
        get() = this.id!! % 2 == 0L
}
