package dev.fyloz.colorrecipesexplorer.service

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.exception.EntityNotFoundException
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.repository.RecipeRepository
import dev.fyloz.colorrecipesexplorer.service.files.FileService
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import org.springframework.mock.web.MockMultipartFile
import java.io.File
import java.nio.file.NoSuchFileException
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class RecipeServiceTest :
    AbstractExternalModelServiceTest<Recipe, RecipeSaveDto, RecipeUpdateDto, RecipeService, RecipeRepository>() {
    override val repository: RecipeRepository = mock()
    private val companyService: CompanyService = mock()
    private val mixService: MixService = mock()
    private val groupService: EmployeeGroupService = mock()
    override val service: RecipeService = spy(RecipeServiceImpl(repository, companyService, mixService, groupService))

    private val company: Company = company(id = 0L)
    override val entity: Recipe = recipe(id = 0L, name = "recipe", company = company)
    override val anotherEntity: Recipe = recipe(id = 1L, name = "another recipe", company = company)
    override val entitySaveDto: RecipeSaveDto = spy(recipeSaveDto(name = entity.name, companyId = entity.company.id!!))
    override val entityUpdateDto: RecipeUpdateDto = spy(recipeUpdateDto(id = entity.id!!, name = entity.name))

    @AfterEach
    override fun afterEach() {
        reset(companyService, mixService)
        super.afterEach()
    }

    // existsByCompany()

    @Test
    fun `existsByCompany() returns true when at least one recipe exists for the given company`() {
        whenever(repository.existsByCompany(company)).doReturn(true)

        val found = service.existsByCompany(company)

        assertTrue(found)
    }

    @Test
    fun `existsByCompany() returns false when no recipe exists for the given company`() {
        whenever(repository.existsByCompany(company)).doReturn(false)

        val found = service.existsByCompany(company)

        assertFalse(found)
    }

    // getAllByCompany()

    @Test
    fun `getAllByCompany() returns the recipes with the given company`() {
        val companies = listOf(entity, anotherEntity)
        whenever(repository.findAllByCompany(company)).doReturn(companies)

        val found = service.getAllByCompany(company)

        assertEquals(companies, found)
    }

    // save()

    @Test
    override fun `save(dto) calls and returns save() with the created entity`() {
        whenever(companyService.getById(company.id!!)).doReturn(company)
        withBaseSaveDtoTest(entity, entitySaveDto, service, { argThat { this.id == null && this.color == color } })
    }

    // update()

    @Test
    override fun `update(dto) calls and returns update() with the created entity`() =
        withBaseUpdateDtoTest(entity, entityUpdateDto, service, { any() })

    // updatePublicData()

    @Test
    fun `updatePublicData() updates the notes of a recipe groups information according to the RecipePublicDataDto`() {
        val recipe = recipe(
            id = 0L, groupsInformation = setOf(
                recipeGroupInformation(id = 0L, group = employeeGroup(id = 1L), note = "Old note"),
                recipeGroupInformation(id = 1L, group = employeeGroup(id = 2L), note = "Another note"),
                recipeGroupInformation(id = 2L, group = employeeGroup(id = 3L), note = "Up to date note")
            )
        )
        val notes = setOf(
            noteDto(groupId = 1, content = "Note 1"),
            noteDto(groupId = 2, content = null)
        )
        val publicData = recipePublicDataDto(recipeId = recipe.id!!, notes = notes)

        doReturn(recipe).whenever(service).getById(recipe.id!!)
        doAnswer { it.arguments[0] }.whenever(service).update(any<Recipe>())

        service.updatePublicData(publicData)

        verify(service).update(argThat<Recipe> {
            assertTrue { this.groupsInformation.first { it.group.id == 1L }.note == notes.first { it.groupId == 1L }.content }
            assertTrue { this.groupsInformation.first { it.group.id == 2L }.note == notes.first { it.groupId == 2L }.content }
            assertTrue { this.groupsInformation.any { it.group.id == 3L } && this.groupsInformation.first { it.group.id == 3L }.note == null }
            true
        })
        verify(mixService, times(0)).updateLocations(any())
    }

    @Test
    fun `updatePublicData() update the location of a recipe mixes in the mix service according to the RecipePublicDataDto`() {
        val publicData = recipePublicDataDto(mixesLocation = setOf(
            mixLocationDto(mixId = 0L, location = "Loc 1"),
            mixLocationDto(mixId = 1L, location = "Loc 2")
        ))

        service.updatePublicData(publicData)

        verify(mixService).updateLocations(publicData.mixesLocation!!)
        verify(service, times(0)).update(any<Recipe>())
    }

    // addMix()

    @Test
    fun `addMix() adds the given mix to the given recipe and updates it`() {
        val mix = mix(id = 0L)
        val recipe = recipe(id = 0L, mixes = mutableListOf())

        doAnswer { it.arguments[0] }.whenever(service).update(any<Recipe>())

        val found = service.addMix(recipe, mix)

        verify(service).update(any<Recipe>())

        assertEquals(recipe.id, found.id)
        assertTrue(found.mixes.contains(mix))
    }

    // removeMix()

    @Test
    fun `removeMix() removes the given mix from its recipe and updates it`() {
        val recipe = recipe(id = 0L, mixes = mutableListOf())
        val mix = mix(id = 0L, recipe = recipe)
        recipe.mixes.add(mix)

        doAnswer { it.arguments[0] }.whenever(service).update(any<Recipe>())

        val found = service.removeMix(mix)

        verify(service).update(any<Recipe>())

        assertEquals(recipe.id, found.id)
        assertFalse(found.mixes.contains(mix))
    }
}

class RecipeImageServiceTest {
    private val recipeService: RecipeService = mock()
    private val fileService: FileService = mock()
    private val service = spy(RecipeImageServiceImpl(recipeService, fileService))

    private val recipeId = 1L
    private val imageId = 5L
    private val imagePath = "$RECIPE_IMAGES_DIRECTORY/$recipeId/$imageId"
    private val recipe = recipe(id = recipeId)
    private val recipeDirectory: File = mock()
    private val imagesIds = listOf(1L, 3L, 10L, 21L)
    private val imageData = byteArrayOf(64, 32, 16, 8, 4, 2, 1)
    private val image = MockMultipartFile("$imageId", imageData)

    @AfterEach
    internal fun tearDown() {
        reset(recipeService, fileService, service, recipeDirectory)
    }

    // getByIdForRecipe()

    @Test
    fun `getByIdForRecipe() returns data for the given recipe and image id red by the file service`() {
        whenever(fileService.getPath(imagePath)).doReturn(imagePath)
        whenever(fileService.readAsBytes(imagePath)).doReturn(imageData)

        val found = service.getByIdForRecipe(imageId, recipeId)

        assertEquals(imageData, found)
    }

    @Test
    fun `getByIdForRecipe() throws EntityNotFoundException when no image with the given recipe and image id exists`() {
        doReturn(imagePath).whenever(service).getPath(imageId, recipeId)
        whenever(fileService.readAsBytes(imagePath)).doAnswer { throw NoSuchFileException(imagePath) }

        val exception =
            assertThrows<EntityNotFoundException> { service.getByIdForRecipe(imageId, recipeId) }
        assertEquals("$recipeId/$imageId", exception.value)
    }


    // getAllIdsForRecipe()

    @Test
    fun `getAllIdsForRecipe() returns a list containing all image's identifier of the images of the given recipe`() {
        val expectedFiles = imagesIds.map { File(it.toString()) }.toTypedArray()

        whenever(recipeService.getById(recipeId)).doReturn(recipe)
        whenever(recipeDirectory.exists()).doReturn(true)
        whenever(recipeDirectory.isDirectory).doReturn(true)
        whenever(recipeDirectory.listFiles()).doReturn(expectedFiles)
        doReturn(recipeDirectory).whenever(service).getRecipeDirectory(recipeId)

        val found = service.getAllIdsForRecipe(recipeId)

        assertEquals(imagesIds, found)
    }

    @Test
    fun `getAllIdsForRecipe() returns an empty list when the given recipe's directory does not exists`() {
        whenever(recipeService.getById(recipeId)).doReturn(recipe)
        whenever(recipeDirectory.exists()).doReturn(false)
        whenever(recipeDirectory.isDirectory).doReturn(true)
        doReturn(recipeDirectory).whenever(service).getRecipeDirectory(recipeId)

        val found = service.getAllIdsForRecipe(recipeId)

        assertTrue(found.isEmpty())
    }

    @Test
    fun `getAllIdsForRecipe() returns an empty list when the given recipe's directory is not a directory`() {
        whenever(recipeService.getById(recipeId)).doReturn(recipe)
        whenever(recipeDirectory.exists()).doReturn(true)
        whenever(recipeDirectory.isDirectory).doReturn(false)
        doReturn(recipeDirectory).whenever(service).getRecipeDirectory(recipeId)

        val found = service.getAllIdsForRecipe(recipeId)

        assertTrue(found.isEmpty())
    }

    // save()

    @Test
    fun `save() writes the given image to the file service with the expected path`() {
        val expectedNextAvailableId = imagesIds.maxOrNull()!! + 1
        val imagePath = "$RECIPE_IMAGES_DIRECTORY/$recipeId/$expectedNextAvailableId"

        doReturn(imagesIds).whenever(service).getAllIdsForRecipe(recipeId)
        doReturn(imagePath).whenever(service).getPath(expectedNextAvailableId, recipeId)

        service.save(image, recipeId)

        verify(fileService).write(image, imagePath)
    }

    // delete()

    @Test
    fun `delete() deletes the image with the given recipe and image id from the file service`() {
        doReturn(imagePath).whenever(service).getPath(imageId, recipeId)

        service.delete(imageId, recipeId)

        verify(fileService).delete(imagePath)
    }

    // getRecipeDirectory()

    @Test
    fun `getRecipeDirectory() returns a file with the expected path`() {
        val recipeDirectoryPath = "$RECIPE_IMAGES_DIRECTORY/$recipeId"
        whenever(fileService.getPath(recipeDirectoryPath)).doReturn(recipeDirectoryPath)

        val found = service.getRecipeDirectory(recipeId)

        assertEquals(recipeDirectoryPath, found.path)
    }

    // getPath()

    @Test
    fun `getPath() returns the expected path`() {
        whenever(fileService.getPath(any())).doAnswer { it.arguments[0] as String }

        val found = service.getPath(imageId, recipeId)

        assertEquals(imagePath, found)
    }
}
